DEFAULT_QUERY_TIMEOUT = 180
MAX_PAGE_SIZE = 1000  # arbitrary large number to return all results

API_KEY_CONFIG_KEY = "api_key"
BEARER_TOKEN_CONFIG_KEY = "bearer_token"
BEARER_TOKEN_EXPIRES_AT_CONFIG_KEY = "bearer_token_expires_at"
MQL_OVERRIDE_CONFIG_KEY = "mql_server_url_override"
PINNED_MODEL_ID_CONFIG_KEY = "pinned_model_id"
